"""DI Dataset Parameter Type."""
from typing import Optional

from cmem.cmempy.workspace.search import list_items
from cmem_plugin_base.dataintegration.types import StringParameterType, Autocompletion
from cmem_plugin_base.dataintegration.utils import setup_cmempy_super_user_access


class DatasetParameterType(StringParameterType):
    """Dataset parameter type."""

    allow_only_autocompleted_values: bool = True

    autocomplete_value_with_labels: bool = True

    dataset_type: Optional[str] = None

    def __init__(
        self, dataset_type: str = None
    ):
        """Dataset parameter type."""
        self.dataset_type = dataset_type

    def autocomplete(
        self, query_terms: list[str], project_id: Optional[str] = None
    ) -> list[Autocompletion]:
        setup_cmempy_super_user_access()
        datasets = list_items(
            item_type="dataset",
            project=project_id
        )["results"]

        result = []
        for _ in datasets:
            identifier = _["id"]
            title = _["label"]
            label = f"{title} ({identifier})"
            if self.dataset_type is not None and self.dataset_type != _["pluginId"]:
                # Ignore datasets of other types
                continue
            for term in query_terms:
                if term.lower() in label.lower():
                    result.append(Autocompletion(value=identifier, label=label))
                    continue
            if len(query_terms) == 0:
                # add any dataset to list if no search terms are given
                result.append(Autocompletion(value=identifier, label=label))
        result.sort(key=lambda x: x.label)  # type: ignore
        return list(set(result))



"""Utils for dataintegration plugins."""
import os
import re


def generate_id(name: str) -> str:
    """Generates a valid DataIntegration identifier from a string.
    Characters that are not allowed in an identifier are removed.
    """
    return re.sub(r"[^a-zA-Z0-9_-]", "", name)


def setup_cmempy_super_user_access():
    """Setup environment for accessing CMEM with cmempy.

    The helper function is used to setup the environment for accessing CMEM with cmempy.
    It does nothing if there is already a working environment.
    If not, it will try to use the configured DI environment.
    """
    try:
        os.environ["OAUTH_GRANT_TYPE"] = "client_credentials"
        if "CMEM_BASE_URI" not in os.environ:
            os.environ["CMEM_BASE_URI"] = os.environ[
                "DEPLOY_BASE_URL"
            ]
        if "OAUTH_CLIENT_ID" not in os.environ:
            os.environ["OAUTH_CLIENT_ID"] = os.environ[
                "DATAINTEGRATION_CMEM_SERVICE_CLIENT"
            ]
        if "OAUTH_CLIENT_SECRET" not in os.environ:
            os.environ["OAUTH_CLIENT_SECRET"] = os.environ[
                "DATAINTEGRATION_CMEM_SERVICE_CLIENT_SECRET"
            ]
    except KeyError as error:
        raise ValueError("Super user configuration not available.") from error



"""All Plugins base classes."""
import logging
from typing import Sequence, Optional

from .entity import Entities


class PluginLogger:
    """Logging API for Plugins.
    If a plugin is running within DataIntegration, this class
    will be replaced to log into DI using the path: plugins.python.<plugin_id>."""

    # pylint: disable=no-self-use

    def debug(self, message: str) -> None:
        """Log a message with severity 'DEBUG'."""
        logging.debug(message)

    def info(self, message: str) -> None:
        """Log a message with severity 'INFO'."""
        logging.info(message)

    def warning(self, message: str) -> None:
        """Log a message with severity 'WARNING'."""
        logging.warning(message)

    def error(self, message: str) -> None:
        """Log a message with severity 'ERROR'."""
        logging.error(message)


class PluginConfig:
    """Configuration API for Plugins.
    If a plugin is running within DataIntegration,
    this class will be replaced to retrieve the DI configuration
    in the path: plugins.python.<plugin_id>."""

    # pylint: disable=no-self-use

    def get(self) -> str:
        """Retrieve plugin configuration as a JSON string.
        This test implementation will return an empty string."""
        return ""


class PluginBase:
    """Base class of all plugins."""

    log: PluginLogger = PluginLogger()

    config: PluginConfig = PluginConfig()


class WorkflowPlugin(PluginBase):
    """Base class of all workflow operator plugins."""

    def execute(self, inputs: Sequence[Entities]) -> Optional[Entities]:
        """Executes the workflow plugin on a given collection of entities.

        :param inputs: Contains a separate collection of entities for each
            input. Currently, DI sends ALWAYS an input. in case no connected
            input is there, the sequence has a length of 0.

        :return: The entities generated from the inputs. At the moment, only one
            entities objects be returned (means only one outgoing connection)
            or none (no outgoing connection).
        """


class TransformPlugin(PluginBase):
    """
    Base class of all transform operator plugins.
    """

    def transform(self, inputs: Sequence[Sequence[str]]) -> Sequence[str]:
        """
        Transforms a collection of values.
        :param inputs: A sequence which contains as many elements as there are input
            operators for this transformation.
            For each input operator it contains a sequence of values.
        :return: The transformed values.
        """



"""Instance of any given concept."""
from typing import Sequence, Iterator


class EntityPath:
    """A path in a schema.

    :param path: The path string using the Silk path language.
    """

    def __init__(self, path: str) -> None:
        self.path = path


class EntitySchema:
    """An entity schema.

    :param type_uri: The entity type
    :param paths: Ordered list of paths
    """

    def __init__(self, type_uri: str, paths: Sequence[EntityPath]) -> None:
        self.type_uri = type_uri
        self.paths = paths


class Entity:
    """An Entity can represent an instance of any given concept.

    :param uri: The URI of this entity
    :param values: All values of this entity. Contains a sequence of values for
        each path in the schema.

    TODO: uri generation
    """

    def __init__(self, uri: str, values: Sequence[Sequence[str]]) -> None:
        self.uri = uri
        self.values = values


class Entities:
    """Holds a collection of entities and their schema.

    :param entities: An iterable collection of entities. May be very large, so it
        should be iterated over and not loaded into memory at once.
    :param schema: All entities conform to this entity schema.
    """

    def __init__(self, entities: Iterator[Entity], schema: EntitySchema) -> None:
        self.entities = entities
        self.schema = schema



"""Parameter types."""
from dataclasses import dataclass
from enum import Enum
from inspect import Parameter
from typing import Optional, TypeVar, Generic, Type, Iterable


@dataclass(frozen=True, eq=True)
class Autocompletion:
    """A single auto-completion result."""

    value: str
    """The value to which the parameter value should be set."""

    label: Optional[str]
    """An optional label that a human user would see instead."""


T = TypeVar("T")


class ParameterType(Generic[T]):
    """Represents a plugin parameter type.
    Provides string-based serialization and autocompletion."""

    name: str
    """The name by which this type can be identified. If available,
    this should be the same as the corresponding DataIntegration type name."""

    allow_only_autocompleted_values: bool = False
    """Hint to the UI that it should only allow to set values for the parameter coming
     from the auto-completion.
    """

    autocomplete_value_with_labels: bool = False
    """Signals that the auto-completed values have labels that must be
    displayed to the user."""

    # flake8: noqa
    # pylint: disable=no-member
    def get_type(self):
        """Retrieves the type that is supported by a given instance."""
        return self.__orig_bases__[0].__args__[0]

    def from_string(self, value: str) -> T:
        """Parses strings into parameter values."""

    # pylint: disable=no-self-use
    def to_string(self, value: T) -> str:
        """Converts parameter values into their string representation."""
        return str(value)

    # pylint: disable=unused-argument
    def autocomplete(
        self, query_terms: list[str], project_id: Optional[str] = None
    ) -> list[Autocompletion]:
        """Autocompletion request.
        Returns all results that match ALL provided query terms.

        :param query_terms: A list of lower case conjunctive search terms.
        :param project_id: The identifier of the project.
        """
        return []

    def label(self, value: str, project_id: Optional[str] = None) -> Optional[str]:
        """Returns the label if exists for the given value."""
        return None

    def autocompletion_enabled(self) -> bool:
        """True, if autocompletion should be enabled on this type.
        By default, checks if the type implements its own autocomplete method."""
        return type(self).autocomplete != ParameterType.autocomplete


class StringParameterType(ParameterType[str]):
    """String type"""

    name = "string"

    def from_string(self, value: str) -> str:
        return value


class IntParameterType(ParameterType[int]):
    """Int type"""

    name = "Long"

    def from_string(self, value: str) -> int:
        return int(value)


class FloatParameterType(ParameterType[float]):
    """Float type"""

    name = "double"

    def from_string(self, value: str) -> float:
        return float(value)


class BoolParameterType(ParameterType[bool]):
    """Boolean type"""

    name = "boolean"

    def from_string(self, value: str) -> bool:
        lower = value.lower()
        if lower in ("true", "1"):
            return True
        if lower in ("false", "0"):
            return False
        raise ValueError("Value must be either 'true' or 'false'")

    def to_string(self, value: bool) -> str:
        if value:
            return "true"
        return "false"


class EnumParameterType(ParameterType[Enum]):
    """Enumeration type"""

    name = "enumeration"

    allow_only_autocompleted_values = True

    def __init__(self, enum_type: Type[Enum]):
        super().__init__()
        self.enum_type = enum_type

    def from_string(self, value: str) -> Enum:
        values = self.enum_type.__members__
        if not value:
            raise ValueError("Empty value is not allowed.")
        if value not in values:
            vals = ", ".join(list(values.keys()))
            raise ValueError(f"'{value}' is not a valid value. Valid values: {vals}.")
        return values[value]

    def to_string(self, value: Enum) -> str:
        return value.name

    def autocomplete(
        self, query_terms: list[str], project_id: Optional[str] = None
    ) -> list[Autocompletion]:
        values = self.enum_type.__members__.keys()
        return list(self.find_matches(query_terms, values))

    @staticmethod
    def find_matches(
        lower_case_terms: list[str], values: Iterable[str]
    ) -> Iterable[Autocompletion]:
        """Finds autocompletions in a list of values"""
        for value in values:
            if EnumParameterType.matches_search_term(lower_case_terms, value.lower()):
                yield Autocompletion(value, value)

    @staticmethod
    def matches_search_term(lower_case_terms: list[str], search_in: str) -> bool:
        """Tests if a string contains a list of (lower case) search terms."""
        lower_case_text = search_in.lower()
        return all(search_term in lower_case_text for search_term in lower_case_terms)


basic_types: list[ParameterType] = [
    StringParameterType(),
    BoolParameterType(),
    IntParameterType(),
    FloatParameterType(),
]


def get_type(param_type: Type) -> ParameterType:
    """Retrieves the ParameterType instance for a given type."""

    if issubclass(param_type, Enum):
        return EnumParameterType(param_type)
    found_type = next(
        (t for t in basic_types if issubclass(param_type, t.get_type())), None
    )
    if found_type is None:
        mapped = map(lambda t: str(t.get_type().__name__), basic_types)
        raise ValueError(
            f"Parameter has an unsupported type {param_type.__name__}. "
            "Supported types are: Enum, "
            f"{', '.join(list(mapped))}."
        )
    return found_type


def get_param_type(param: Parameter) -> ParameterType:
    """Retrieves the ParameterType instance for a given parameter."""

    if param.annotation == Parameter.empty:
        # If there is no type annotation, DI should send the parameter as a string
        return StringParameterType()
    return get_type(param.annotation)



"""Knowledge Graph Parameter Type."""
from typing import Optional, Set, List

from cmem.cmempy.dp.proxy.graph import get_graphs_list

from cmem_plugin_base.dataintegration.types import StringParameterType, Autocompletion
from cmem_plugin_base.dataintegration.utils import setup_cmempy_super_user_access


class GraphParameterType(StringParameterType):
    """Knowledge Graph parameter type."""

    allow_only_autocompleted_values: bool = False

    autocomplete_value_with_labels: bool = True

    classes: Optional[Set[str]] = None

    def __init__(
        self,
        show_di_graphs: bool = False,
        show_system_graphs: bool = False,
        show_graphs_without_class: bool = False,
        classes: List[str] = None,
        allow_only_autocompleted_values: bool = True,
    ):
        """
        Knowledge Graph parameter type.

        :param show_di_graphs: show DI project graphs
        :param show_system_graphs: show system graphs such as shape and query catalogs
        :param classes: allowed classes of the shown graphs
            - if None -> defaults to di:Dataset and void:Dataset
        :param allow_only_autocompleted_values: allow entering new graph URLs
        """
        self.show_di_graphs = show_di_graphs
        self.show_system_graphs = show_system_graphs
        self.show_graphs_without_class = show_graphs_without_class
        self.allow_only_autocompleted_values = allow_only_autocompleted_values
        if classes:
            self.classes = set(classes)
        else:
            self.classes = {
                "https://vocab.eccenca.com/di/Dataset",
                "http://rdfs.org/ns/void#Dataset"
            }

    def autocomplete(
        self, query_terms: list[str], project_id: Optional[str] = None
    ) -> list[Autocompletion]:
        setup_cmempy_super_user_access()
        graphs = get_graphs_list()
        result = []
        for _ in graphs:
            iri = _["iri"]
            title = _["label"]["title"]
            label = f"{title} ({iri})"
            assigned_classes = set(_["assignedClasses"])
            # ignore DI project graphs
            if self.show_di_graphs is False and _["diProjectGraph"] is True:
                continue
            # ignore system resource graphs
            if self.show_system_graphs is False and _["systemResource"] is True:
                continue
            # show graphs without assigned classes only if explicitly wanted
            if len(assigned_classes) == 0:
                if self.show_graphs_without_class is True:
                    result.append(Autocompletion(value=iri, label=label))
                continue
            # ignore graphs which do not match the requested classes
            if (
                self.classes is not None
                and len(assigned_classes) > 0
                and len(self.classes.intersection(assigned_classes)) == 0
            ):
                continue
            # if no search terms are given: add all remaining graphs to list
            if len(query_terms) == 0:
                result.append(Autocompletion(value=iri, label=label))
                continue
            # show only graphs which match the given terms
            for term in query_terms:
                if term.lower() in label.lower():
                    result.append(Autocompletion(value=iri, label=label))
                    continue
        result.sort(key=lambda x: x.label)  # type: ignore
        return list(set(result))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cmem_plugin_base',
 'cmem_plugin_base.dataintegration',
 'cmem_plugin_base.dataintegration.parameter']

package_data = \
{'': ['*']}

install_requires = \
['cmem-cmempy>=21.11.5']

setup_kwargs = {
    'name': 'cmem-plugin-base',
    'version': '1.1.0',
    'description': 'Base classes for developing eccenca Coporate Memory plugins.',
    'long_description': '# cmem-plugin-base\n\nPython base classes for developing eccenca Coporate Memory plugins.\n\nIn order to kick-start developing eccenca Corporate Memory Plugins, please check out this project template: https://github.com/eccenca/cmem-plugin-template\n\n',
    'author': 'eccenca',
    'author_email': 'cmempy-developer@eccenca.com',
    'maintainer': 'Sebastian Tramp',
    'maintainer_email': 'sebastian.tramp@eccenca.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Package and plugin discovery module."""
import importlib
import importlib.util
import json
import pkgutil
import sys
from subprocess import check_output  # nosec
from types import ModuleType
from typing import Sequence

from cmem_plugin_base.dataintegration.description import PluginDescription, Plugin


def get_packages():
    """Get installed python packages.

    Returns a list of dict with the following keys:
     - name - package name
     - version - package version
    """
    return json.loads(check_output(["pip", "list", "--format", "json"], shell=False))


def discover_plugins_in_module(
    package_name: str = "cmem",
) -> Sequence[PluginDescription]:
    """Finds all plugins within a base package.

    :param package_name: The base package. Will recurse into all submodules
        of this package.
    """

    def load_module(module_name: str) -> ModuleType:
        module_is_imported = module_name in sys.modules
        module = importlib.import_module(module_name)
        if module_is_imported:
            importlib.reload(module)  # need to reload in order to discover plugins
        return module

    def import_submodules(module: ModuleType):
        for _loader, name, is_pkg in pkgutil.walk_packages(module.__path__):
            sub_module = load_module(module.__name__ + "." + name)
            if is_pkg:
                import_submodules(sub_module)

    Plugin.plugins = []
    import_submodules(load_module(package_name))
    return Plugin.plugins


def discover_plugins(package_name: str = "cmem_plugin"):
    """Discover plugin descriptions in packages.

    This is the main discovery method which is executed by DataIntegration.
    It will go through all modules which base names starts with
    package_name.

    :param package_name: The package prefix.
    """
    target_packages = []
    plugin_descriptions = []
    # select prefixed packages
    for module in pkgutil.iter_modules():
        name = module.name
        if name.startswith(package_name) and name != "cmem_plugin_base":
            target_packages.append(name)
    for name in target_packages:
        for plugin in discover_plugins_in_module(package_name=name):
            plugin_descriptions.append(plugin)
    return plugin_descriptions



"""Classes for describing plugins"""
import inspect
from inspect import _empty
from typing import Optional, List, Type, Any

from cmem_plugin_base.dataintegration.plugins import WorkflowPlugin, TransformPlugin
from cmem_plugin_base.dataintegration.types import ParameterType, get_param_type
from cmem_plugin_base.dataintegration.utils import generate_id


class PluginParameter:
    """A plugin parameter.

    :param name: The name of the parameter
    :param label: A human-readable label of the parameter
    :param description: A human-readable description of the parameter
    :param param_type: Optionally overrides the parameter type.
        Usually does not have to be set manually as it will be inferred from the
        plugin automatically.
    :param default_value: The parameter default value (optional)
        Will be inferred from the plugin automatically.
    :param advanced: True, if this is an advanced parameter that should only be
        changed by experienced users
    """

    def __init__(
        self,
        name: str,
        label: str = "",
        description: str = "",
        param_type: Optional[ParameterType] = None,
        default_value: Optional[Any] = None,
        advanced: bool = False,
    ) -> None:
        self.name = name
        self.label = label
        self.description = description
        self.param_type = param_type
        self.default_value = default_value
        self.advanced = advanced


class PluginDescription:
    """A plugin description.

    :param plugin_class: The plugin implementation class
    :param label: A human-readable label of the plugin
    :param description: A short (few sentence) description of this plugin.
    :param documentation: Documentation for this plugin in Markdown.
    :param categories: The categories to which this plugin belongs to.
    :param parameters: Available plugin parameters
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        plugin_class,
        label: str,
        plugin_id: str = None,
        description: str = "",
        documentation: str = "",
        categories: List[str] = None,
        parameters: List[PluginParameter] = None,
    ) -> None:
        #  Set the type of the plugin. Same as the class name of the plugin
        #  base class, e.g., 'WorkflowPlugin'.
        if issubclass(plugin_class, WorkflowPlugin):
            self.plugin_type = "WorkflowPlugin"
        elif issubclass(plugin_class, TransformPlugin):
            self.plugin_type = "TransformPlugin"
        else:
            raise ValueError(
                f"Class {plugin_class.__name__} does not implement a supported"
                f"plugin base class (e.g., WorkflowPlugin)."
            )

        self.plugin_class = plugin_class
        self.module_name = plugin_class.__module__
        self.class_name = plugin_class.__name__
        if plugin_id is None:
            self.plugin_id = generate_id(
                (self.module_name + "-" + self.class_name).replace(".", "-")
            )
        else:
            self.plugin_id = plugin_id
        if categories is None:
            self.categories = []
        else:
            self.categories = categories
        self.label = label
        self.description = description
        self.documentation = documentation
        if parameters is None:
            self.parameters = []
        else:
            self.parameters = parameters


class Categories:
    """A list of common plugin categories. At the moment, in the UI,
    categories are only utilized for rule operators, such as transform plugins."""

    # Plugins in the 'Recommended' category will be shown preferably
    RECOMMENDED: str = "Recommended"

    # Common transform categories
    COMBINE: str = "Combine"
    CONDITIONAL: str = "Conditional"
    CONVERSION: str = "Conversion"
    DATE: str = "Date"
    EXCEL: str = "Excel"
    EXTRACT: str = "Extract"
    FILTER: str = "Filter"
    GEO: str = "Geo"
    LINGUISTIC: str = "Linguistic"
    NORMALIZE: str = "Normalize"
    NUMERIC: str = "Numeric"
    PARSER: str = "Parser"
    REPLACE: str = "Replace"
    SCRIPTING: str = "Scripting"
    SELECTION: str = "Selection"
    SEQUENCE: str = "Sequence"
    SUBSTRING: str = "Substring"
    TOKENIZATION: str = "Tokenization"
    VALIDATION: str = "Validation"
    VALUE: str = "Value"


class Plugin:
    """Annotate classes with plugin descriptions.

    :param label: A human-readable label of the plugin
    :param plugin_id: Optionally sets the plugin identifier.
        If not set, an identifier will be generated from the module and class name.
    :param description: A short (few sentence) description of this plugin.
    :param documentation: Documentation for this plugin in Markdown. Note that you
        do not need to add a first level heading to the markdown since the
        documentation rendering component will add a heading anyway.
    :param categories: The categories to which this plugin belongs to.
    :param parameters: Available plugin parameters
    """

    plugins: list[PluginDescription] = []

    def __init__(
        self,
        label: str,
        plugin_id: Optional[str] = None,
        description: str = "",
        documentation: str = "",
        categories: List[str] = None,
        parameters: List[PluginParameter] = None,
    ):
        self.label = label
        self.description = description
        self.documentation = documentation
        self.plugin_id = plugin_id
        if categories is None:
            self.categories = []
        else:
            self.categories = categories
        if parameters is None:
            self.parameters = []
        else:
            self.parameters = parameters

    def __call__(self, func):
        plugin_desc = PluginDescription(
            plugin_class=func,
            label=self.label,
            plugin_id=self.plugin_id,
            description=self.description,
            documentation=self.documentation,
            categories=self.categories,
            parameters=self.retrieve_parameters(func),
        )
        Plugin.plugins.append(plugin_desc)
        return func

    def retrieve_parameters(self, plugin_class: Type) -> List[PluginParameter]:
        """Retrieves parameters from a plugin class and matches them with the user
        parameter definitions."""

        # Only return parameters for user-defined init methods.
        if not hasattr(plugin_class.__init__, "__code__"):
            return []
        # Collect parameters from init method
        params = []
        sig = inspect.signature(plugin_class.__init__)
        for name in sig.parameters:
            if name != "self":
                param = next((p for p in self.parameters if p.name == name), None)
                if param is None:
                    param = PluginParameter(name)
                sig_param = sig.parameters[name]
                if param.param_type is None:
                    param.param_type = get_param_type(sig_param)
                if param.default_value is None and sig_param.default != _empty:
                    param.default_value = sig_param.default
                params.append(param)
        return params
